# -*- coding: utf-8 -*-
'''
DESCRIPTION
   Win2Linux migration linux side
AUTHOR
    Dan Kushner (dan.kushner@sisense.com)
    Valeriy Galakh (valeriy.galakh@sisense.com)
'''
import subprocess
import os
import json
import tarfile
import requests
import argparse
import logging
import time
import inspect
import zipfile
import glob
import shutil
import urllib3
import multiprocessing

urllib3.disable_warnings()

__metaclass__ = type


try:
    raw_input = input
except NameError:
    pass

SAMPLE_CUBES = ['Sample ECommerce', 'Sample Healthcare', 'Sample Lead Generation', 'Sample Retail']

migration_user_hash = {'hash': 'a8228af478d7965505e1e801e3b7e180a84055992513716b8cb2b92c90efe2a42a4bcb75e496b0ed894402d1f6c92e20eaed94b1f55024c0d9aed534acfddeb9',
                       'salt': '8d712102ac054f22db7ed14e756c8c2635f2b02a26450f8def42db0764938d342606884dfd6ac652ba8a9e4bf6e0dbc023aa1f5447e18b433aa675003e6674b3251176ec0a2b793ff856d88c9690b29ad486767f883d3820ea35aa40bc80c2eeec313c288dac5933e68840faa86c0f6f1cb6c2bf626f20fec72a5f0bf6e46ff9',
                       'type': 'SHA256_SALT',
                       'password': 'Sisense1!'}
admin_user = {'hash': None,
              'salt': None,
              'type': None,
              'email': None}

params_dict = dict()


logging.root.handlers = []
log_date = time.strftime("%Y-%m-%d_%H_%M_%S", time.gmtime())
logging.basicConfig(filename='win2linux_sisense_' + str(log_date) + '.log', filemode='w', level=logging.DEBUG, format='[%(asctime)s] - %(levelname)8s ---> %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
# set up logging to console
console = logging.StreamHandler()
console.setLevel(logging.INFO)
# set a format which is simpler for console use
formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(message)s')
console.setFormatter(formatter)
logging.getLogger("").addHandler(console)


class SisenseAPI:
    '''Class for general Sisense API'''

    def __init__(self, username, password, token, linux_base_url, login_urlsuffix):
        self.sisense_base_url_for_login = linux_base_url
        self.login_urlsuffix = login_urlsuffix  # /api/v1/authentication/login

        self.sisense_base_url_ecdata_swagger = linux_base_url

        self.sisense_base_url_import_schema_json = linux_base_url

        self.url_suffix_get_identity = '/api/elasticubes/servers/LocalHost/identity'
        self.url_suffix_version = '/api/v2/version'

        self.token = None
        if username and password:
            logging.info('Trying to use API username and password from config file ...')
            self.token = self.login(username, password)
        if self.token is None:
            logging.info('Try to use existed super user ...')
            if admin_user['email'] is not None:
                self.token = self.login(admin_user['email'], migration_user_hash['password'])
            else:
                self.mongo_conn = MongoDB()
                admin_user['email'], admin_user['hash'], admin_user['salt'], admin_user['type'] = self.mongo_conn.get_user_with_role_id('super')
                if admin_user['email'] is not None:
                    change_pass = self.mongo_conn.change_user_password(admin_user['email'], migration_user_hash['hash'],
                                                                       migration_user_hash['salt'], migration_user_hash['type'])
                    if change_pass:
                        pass_recovery = 'db.users.updateOne({email: "'+admin_user['email']+'"}, {"$set": {"hash.hash": "'+admin_user['hash']+'", "hash.salt": "'+admin_user['salt']+'", "hash.type": "'+admin_user['type']+'"}})'
                        logging.debug('Original password recovery command: {}'.format(self.mongo_conn.db_query('prismWebDB', pass_recovery)))
                        self.token = self.login(admin_user['email'], migration_user_hash['password'])
        if self.token is None:
            logging.info('Trying to use API token from config file ...')
            self.token = token
        if not self.token:
            logging.error('Failed to use Sisense API. Please update username/password or token in config.json')

    def __del__(self):
        if admin_user['email'] is not None:
            retries = 5
            retry_pause = 10
            logging.info('Stop using "{}"'.format(admin_user['email']))
            # Set back old hashes for user (5 retries)
            for i in range(retries):
                password_changed = self.mongo_conn.change_user_password(admin_user['email'], admin_user['hash'],
                                                                        admin_user['salt'], admin_user['type'])
                if password_changed:
                    logging.info('Usage of "{}" is finished'.format(admin_user['email']))
                    for k in admin_user:
                        admin_user[k] = None
                    break
                else:
                    logging.error('Retry in {} seconds'.format(retry_pause))
                    time.sleep(retry_pause)

    @staticmethod
    def call_api(http_method, sisense_base_url, url_suffix, params=None, payload=None, token=None, expected_status=200):
        logging.info('Execute API {0} {1}{2}'.format(http_method, sisense_base_url, url_suffix))
        if token is None:
            headers = {}
        else:
            headers = {'authorization': 'Bearer ' + token}
        if payload:
            headers['Content-Type'] = 'application/json'
        url = sisense_base_url + url_suffix
        response = None
        try:
            if http_method == 'GET':
                response = requests.get(url, params=params, headers=headers, verify=False)
            elif http_method == 'POST':
                data = json.dumps(payload)
                response = requests.post(url, data=data, params=params, headers=headers, verify=False)
            elif http_method == 'DELETE':
                response = requests.delete(url, headers=headers, verify=False)
        except Exception as ex:
            logging.error('Error call api -> {0}'.format(ex))
            return None

        if response.status_code != expected_status or 'error' in response.text:
            logging.info('API call error. URL: {}. Status: {}. Text: {}'.format(response.url, response.status_code,
                                                                                response.text))
        return response

    def login(self, email, password):
        payload = {
            "username": email,
            "password": password
        }
        res = self.call_api('POST', self.sisense_base_url_for_login, self.login_urlsuffix, params=None, payload=payload)
        if res.status_code == 200:
            token = json.loads(res.text)
            logging.info('-Success- connected with {0} via {1}'.format(email, self.login_urlsuffix))
            return token['access_token']
        else:
            logging.error('-Fail- connect with {0} via {1}'.format(email, self.login_urlsuffix))
            return None

    def get_new_linux_server_identity(self):
        logging.info("Getting servers identity")
        try:
            res = self.call_api('GET', self.sisense_base_url_for_login, self.url_suffix_get_identity, params=None,
                                payload=None, token=self.token)
            identity = json.loads(res.text)
            return identity['identity']
        except Exception as e:
            logging.error('Fail: {0} {1} reason {2}'.format(str(e), str(res.text), str(res.reason)))
            logging.info(res)
            return None

    def load_elasticube(self, cube_name):
        url_suffix = '/api/v2/ecm/'
        logging.info(
            "Load elasticube {0} with {1}".format(cube_name, self.sisense_base_url_import_schema_json + url_suffix))
        query = "query loadElasticube($title: String!, $server: String) {\n  elasticubeByTitle(title: $title, server: $server) {\n    oid\n    __typename\n  }\n}\n"
        operation_name = "loadElasticube"
        payload = {
            "query": query,
            "variables": {
                "title": cube_name,
                "server": "LocalHost"
            },
            "operationName": operation_name}
        logging.debug('payload: {}'.format(payload))
        try:
            res = self.call_api('POST', self.sisense_base_url_for_login, url_suffix, params=None,
                                payload=payload, token=self.token)
            if res.status_code != 200:
                logging.error('Fail: {0} reason {1}'.format(str(res.text), str(res.reason)))
                logging.info(res)
        except Exception as ex:
            logging.error('-Fail- execute {0} -> {1}'.format(url_suffix, ex))
            pass

    def get_all_cubes(self):
        url_suffix = '/api/v2/ecm/'
        payload = {"query":"query _ {\n  elasticubesMetadata {\n    oid\n    title\n    server\n    serverId\n    type\n    relationType\n    provider\n    datasets {\n      connection {\n        id\n        autoRefresh\n        refreshRate\n        timeout\n        resultLimit\n        __typename\n      }\n      __typename\n    }\n    set {\n      title\n      __typename\n    }\n    status\n    lastBuildTime\n    lastBuildStatus\n    lastSuccessfulBuildStartTime\n    lastSuccessfulBuildTime\n    lastPublishTime\n    lastUpdated\n    nextBuildTime\n    creator {\n      id\n      __typename\n    }\n    shares {\n      partyId\n      type\n      ... on ShareUserInfo {\n        firstName\n        lastName\n        email\n        __typename\n      }\n      ... on ShareGroupInfo {\n        name\n        ad\n        objectSid\n        everyone\n        admins\n        __typename\n      }\n      __typename\n    }\n    __typename\n  }\n}\n","operationName":"_"}
        result = self.call_api("POST", self.sisense_base_url_for_login, url_suffix, payload=payload, token=self.token)
        if result.status_code == 200:
            return result.json()
        else:
            return None

    def publish_live_elasticube(self, oid):
        url_suffix = '/api/v2/ecm/'
        payload = {
            "operationName": "publishElasticube",
            "query": "mutation publishElasticube($elasticubeOid: UUID!){\npublishElasticube(elasticubeOid: $elasticubeOid)\n}\n",
            "variables": {
                "elasticubeOid": ""
            }
        }
        payload["variables"]["elasticubeOid"] = oid
        result = self.call_api('POST', self.sisense_base_url_for_login, url_suffix, payload=payload, token=self.token)
        return result

    def publish_and_load_cubes(self):
        logging.info("------------------------------------Publish live models------------------------------------")
        all_cubes = self.get_all_cubes()
        if all_cubes is None:
            logging.error('Failed to get cubes oids')
        else:
            for cube in all_cubes['data']['elasticubesMetadata']:
                if cube['title'] not in SAMPLE_CUBES:
                    try:
                        if cube['type'] == 'live':
                            logging.info('Publishing live model {}'.format(cube['title']))
                            self.publish_live_elasticube(cube['oid'])
                        else:
                            logging.info('Loading ec "{}"'.format(cube['title']))
                            self.load_elasticube(cube['title'])
                    except Exception as ex:
                        logging.error('Failed to publish Live model {}. Error: {}'.format(cube['title'], ex))

    def set_branding(self, payload):
        url_suffix = '/api/branding'
        post_result = self.call_api('POST', self.sisense_base_url_for_login, url_suffix, payload=payload, token=self.token)
        if post_result.status_code != 200 or 'error' in post_result.text:
            logging.error('Branding setup error. Status: {}. Text: {}'.format(post_result.status_code, post_result.text))


class Utils:
    """ Class for general utils usage"""

    def __init__(self):
        self.tar_gz_location = os.path.join(os.getcwd(), "sisense_win_assets.tar.gz")
        self.config_attributes = params_dict

    def general_extraction(self):
        """
        :return: assets_path
        """
        if self.config_attributes['extract'] == 'True':
            logging.info("------------------------------------Extract assets------------------------------------")
            assets_path = self.extract_tar_gz()
            self.continue_validation(assets_path, 'extract_tar_gz', False)
            self.permission_and_sudo_usage(assets_path)
        else:
            assets_path = os.path.join(self.config_attributes['working_dir'], 'assets_to_linux')
        logging.info('Assets path: {}'.format(assets_path))
        return assets_path

    def continue_validation(self, method_result, method_name, cancel_script_flag):
        if method_result is None:
            time.sleep(1)
            if cancel_script_flag:
                logging.error('Fail {0}, please contact support'.format(method_name))
                exit(0)
            else:
                self.continue_alert_msg('Fail "{0}" continue to next task?'.format(method_name), True)

    @staticmethod
    def continue_alert_msg(prompt=None, resp=False):
        """prompts for yes or no response from the user. Returns True for yes and
        False for no.user simply types ENTER.
        """
        if resp:
            prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
        else:
            prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')

        while True:
            ans = raw_input(prompt)
            if not ans:
                return resp
            if ans not in ['y', 'Y', 'n', 'N']:
                print('please enter y or n.')
                continue
            if ans == 'y' or ans == 'Y':
                return True
            if ans == 'n' or ans == 'N':
                exit(0)

    def execute_cmd(self, cmd, print_out=True):
        ''' Execute cmd without no response required'''
        try:
            logging.info('Execute-> {0}'.format(cmd))
            output = subprocess.check_output(cmd, shell=True).decode("utf-8")
            if ('Error' in output) or ('Fail' in output):
                logging.error(output)
                return self.exception('Fail')
            else:
                if print_out:
                    logging.info(output)
                else:
                    logging.debug(output)
                return output
        except Exception as ex:
            return self.exception(ex)

    def extract_tar_gz(self):
        try:
            logging.info("Extract sisense_win_assets tar file")
            tar = tarfile.open(self.tar_gz_location)
            for i, m in enumerate(tar.getmembers()):
                if i == 0:
                    path = str(m.name)
                    logging.info("Extracting: " + path)
                tar.extract(m)
            tar.close()
            return path
        except Exception as ex:
            return self.exception(ex)

    def permission_and_sudo_usage(self, assets_path):
        logging.info('Setup a+x permission to ./{0}'.format(str(assets_path)))
        cmd_permission = 'sudo chmod a+x -R ./{0}'.format(str(assets_path))
        self.execute_cmd(cmd_permission)
        cmd_option_to_run_as_sudo = 'sudo cp -fr /usr/local/bin/kubectl /usr/bin/'
        self.execute_cmd(cmd_option_to_run_as_sudo)
        return True

    def api_authurization(self):
        api = SisenseAPI(self.config_attributes['api_username'], self.config_attributes['api_password'], self.config_attributes['api_token'],
                         self.config_attributes['linux_base_url'], self.config_attributes['login_urlsuffix'])
        return api

    @staticmethod
    def exception(ex):
        logging.error("Fail -> {0} Error: {1}  ".format(str(inspect.stack()[1][3]), str(ex)))
        raw_input("Press Enter to continue next step...")
        return None

    @staticmethod
    def create_folder(folder_path, delete_if_exists=True):
        try:
            if not os.path.splitext(folder_path)[1]:
                if os.path.exists(folder_path):
                    if delete_if_exists:
                        # remove
                        logging.info('Deleting folder {}'.format(folder_path))
                        shutil.rmtree(folder_path, ignore_errors=True)
                    else:
                        return
                # create
                logging.info('Creating folder {}'.format(folder_path))
                os.makedirs(folder_path)
            else:
                raise Exception('"{}" is not a folder path'.format(folder_path))
        except Exception as e:
            logging.error('Failed to create folder "{}"'.format(folder_path))
            raise e


class K8S(Utils):
    '''
    Class k8s usage (All kind of k8s commands usage and flow)
    '''

    def __init__(self):
        self.kubectl_call = 'kubectl --kubeconfig=$HOME/.kube/config'
        super(K8S, self).__init__()

    # ToDo  use get pods -l app='identity' -o=name. Adapt for cluster and get primary pod
    def get_pod_name(self, pod, namespace):
        try:
            get_pod = '{0} -n {1} get pods -o=name | grep {2}'.format(self.kubectl_call, namespace, pod)
            logging.info('get_pod_name_cmd -> {0}'.format(get_pod))
            pod_name = subprocess.check_output(get_pod, shell=True).decode('utf-8').split('/')[-1]
            logging.info('Pod name: {0}'.format(pod_name))
            return pod_name
        except Exception as ex:
            return self.exception(ex)

    def copy_to_pod(self, file_copy, namespace, pod_name, pod_dest_dir):
        copy_cmd = '{0} cp "{1}" {2}/{3}:{4}'.format(self.kubectl_call, file_copy, namespace, pod_name, pod_dest_dir)
        self.continue_validation(self.execute_cmd(copy_cmd, print_out=False), 'copy_cmd', False)

    def copy_ecdata_files_to_destenation(self, assets_path):
        logging.info("------------------------------------Copy ecdata------------------------------------")
        try:
            ignore_cubes = ['Usage Analytics Model.ecdata']
            path = os.path.join(self.config_attributes['working_dir'], assets_path)
            ecdata_files = [f for f in os.listdir(path) if f.endswith('.ecdata') and f not in ignore_cubes]
            if len(ecdata_files) == 0:
                logging.info('No ecdata files found.')
                return True
            logging.info(ecdata_files)
            managment_pod_name = self.get_pod_name('management', self.config_attributes['managment_namespace']).rstrip()
            logging.info("--------------------------Copy ecdata to pod storage location--------------------------")
            for ecdata in ecdata_files:
                try:
                    ecdata_full_path = os.path.join(path, ecdata)
                    logging.info("Copy {0} to storage location: {1}".format(ecdata, self.config_attributes['ecdata_storage_path']))
                    self.copy_to_pod(ecdata_full_path, self.config_attributes['managment_namespace'], managment_pod_name,
                                     self.config_attributes['ecdata_storage_path'])
                    logging.info("Removing {0} from: {1}".format(ecdata, path))
                    os.remove(ecdata_full_path)
                except Exception as ex:
                    logging.error('Error copy_ecdata_files_to_destenation -> {0}'.format(ex))
                    pass
            return True
        except Exception as ex:
            return self.exception(ex)

    def get_current_plugin_list(self, plugin_pod_name):
        get_plugins_cmd = '{0} -n {1} exec -i {2} -- ls {3}'.format(self.kubectl_call,
                                                                    self.config_attributes['managment_namespace'],
                                                                    plugin_pod_name, self.config_attributes['plugins'])
        list_of_plugins = subprocess.check_output(get_plugins_cmd, shell=True).decode('utf-8').splitlines()
        logging.info(list_of_plugins)
        return list_of_plugins

    def copy_plugins_to_destination(self, assets_path):
        try:
            logging.info("--------------------------Copy plugins to pod location--------------------------")
            plugins_pod_name = self.get_pod_name('plugins', self.config_attributes['managment_namespace']).rstrip()
            path = os.path.join(self.config_attributes['working_dir'], assets_path + '/plugins')
            plugins_current_list = self.get_current_plugin_list(plugins_pod_name)
            all_plugins = os.listdir(path)
            for plugin in all_plugins:
                if plugin not in plugins_current_list:
                    logging.info("Copy {0} to storage location".format(plugin))
                    self.copy_to_pod(str(os.path.join(path, plugin)), self.config_attributes['managment_namespace'],
                                     plugins_pod_name, self.config_attributes['plugins'])
            self.delete_pod(self.get_pod_name('plugins', self.config_attributes['managment_namespace']).rstrip())
            return True
        except Exception as ex:
            return self.exception(ex)

    def copier_worker(self, params_tuple):
        self.copy_to_pod(params_tuple[0], params_tuple[1], params_tuple[2], params_tuple[3])

    def copy_static_assets_to_destination(self, assets_path):
        def replace_branding_path(data):
            if type(data) is list:
                return [replace_branding_path(item) for item in data]

            if type(data) is dict:
                return {key: replace_branding_path(value) for key, value in data.items()}

            if type(data) is str:
                if data == 'false':
                    return False
                if data == 'true':
                    return True
                data = data.replace('\\', '/')
                data = data.replace('resources/branding', 'branding')
                return data

        logging.info('--------------------------Copy static assets--------------------------')
        try:
            managment_pod_name = self.get_pod_name('management', self.config_attributes['managment_namespace']).rstrip()
            for k, v in self.config_attributes['static_assets_paths'].items():
                path = os.path.join(self.config_attributes['working_dir'], assets_path, k, '*')
                logging.info(">>>>>>>>>>> Copy {0} to storage location -> {1}".format(path, v))
                params = [(f, self.config_attributes['managment_namespace'], managment_pod_name, v) for f in glob.glob(path)]
                with multiprocessing.Pool(10) as p:
                    p.map(self.copier_worker, params)

            # Branding migration
            # Read zk configurations configurationExport.json
            logging.info('--------------------------Check branding--------------------------')
            zk_conf_file = os.path.join(self.config_attributes['working_dir'], assets_path, 'zookeeper/configurationExport.json')
            if os.path.exists(zk_conf_file):
                try:
                    with open(zk_conf_file, 'r') as f:
                        zk_conf = json.load(f)
                        branding_info = zk_conf['configuration-service']['branding']
                    if branding_info['enabled'] == 'true':
                        logging.info('Activating branding')
                        # Change files paths for Linux
                        branding_changed = replace_branding_path(branding_info)
                        # Set default branding values for nulls
                        if branding_changed['homePageConfig']['hideSearchInNavver'] == 'null':
                            branding_changed['homePageConfig']['hideSearchInNavver'] = True
                        if branding_changed['homePageConfig']['hideMarketingBanner'] == 'null':
                            branding_changed['homePageConfig']['hideMarketingBanner'] = False
                        if branding_changed['poweredBySisense'] == 'null':
                            branding_changed['poweredBySisense'] = True
                        # Setup branding
                        api = self.api_authurization()
                        api.set_branding(branding_changed)
                except Exception as e:
                    logging.warning('Cannot update branding. Error: {}'.format(e))

            else:
                logging.warning('Cannot update branding. configurationExport.json file not found.')

            return True
        except Exception as ex:
            return self.exception(ex)

    def extract_local_datasources(self, input_files_zip, input_files_path):
        logging.info("------------------------------------Extract local datasources------------------------------------")
        # Delete input files dir if exists
        if os.path.exists(input_files_path):
            logging.info('Deleting folder {}'.format(input_files_path))
            shutil.rmtree(input_files_path, ignore_errors=True)
        logging.info('Extract {} into {}'.format(input_files_zip, input_files_path))
        managment_pod_name = self.get_pod_name('management', self.config_attributes['managment_namespace']).rstrip()
        # Create temp dir for unzip
        temp_dir = os.path.join(os.getcwd(), 'temp_input_files')
        self.create_folder(temp_dir)

        # Unzip archive and copy to the destination
        with zipfile.ZipFile(input_files_zip, 'r') as zip_ref:
            logging.info('Extracting archive {} into {}'.format(input_files_zip, temp_dir))
            zip_ref.extractall(temp_dir)
        self.copy_to_pod(temp_dir, self.config_attributes['managment_namespace'], managment_pod_name, input_files_path)

        # Delete temp dir
        logging.info('Deleting folder {}'.format(temp_dir))
        shutil.rmtree(temp_dir, ignore_errors=True)

    def mongodb_restore(self, assets_path):
        mongodb_pod_name, mongodb_container_name = self.mongodb_single_or_cluster_get_pods_name()
        logging.info('Running MongoDB dump and restore')
        try:
            logging.info("------------------------------------Mongodb dump------------------------------------")
            self.continue_validation(self.mongodb_backup(mongodb_pod_name, mongodb_container_name), 'backup mongodb', True)
            logging.info("------------------------------------Mongodb restore------------------------------------")
            logging.info('Copy mongodb dump to pod')
            mongo_db_folder_path = os.path.join(self.config_attributes['working_dir'], assets_path + '/mongo_dump')
            # https://devops.stackexchange.com/questions/3127/how-to-copy-files-recursively-to-a-kubernetes-pod
            mongo_db_folder_copy_to_pod = '{0} cp -n {1} -c {2} {3} {4}:/tmp'.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name,
                       mongo_db_folder_path, mongodb_pod_name)
            self.continue_validation(self.execute_cmd(mongo_db_folder_copy_to_pod), 'mongo_db_folder_copy_to_pod', True)
            for db in os.listdir(mongo_db_folder_path):
                if 'admin' in db:
                    continue
                else:
                    logging.info('drop database {0}'.format(db))
                    drop_mongodb_db_cmd = '{0} -n {1} -c {2} exec -i {3} -- mongo {4} --eval "db.dropDatabase()"'.\
                        format(self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name, mongodb_pod_name, db)
                    self.continue_validation(self.execute_cmd(drop_mongodb_db_cmd), 'drop_mongodb_db_cmd', True)
            cmd_mongorestore = '{0} -n {1} -c {2} exec -i {3} -- mongorestore --drop /tmp/mongo_dump --nsExclude={4} --nsExclude={5} --nsExclude={6}'.format(
                self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name, mongodb_pod_name, 'prismWebDB.elasticubeBuildLogs',
                'prismWebDB.trace', 'prismWebDB.elasticubeRevisions')
            self.continue_validation(self.execute_cmd(cmd_mongorestore), 'cmd_mongorestore', True)
            logging.info('Linux MongoDB deploy done')
            self.delete_pod(self.get_pod_name('api-gateway', self.config_attributes['managment_namespace']).rstrip())
            time.sleep(10)
            self.delete_pod(self.get_pod_name('model-graphql', self.config_attributes['managment_namespace']).rstrip())
            time.sleep(10)

            # Restart migration job
            self.migration_job_update()
            logging.info("Sleeping for 2 minutes...")
            time.sleep(120)

            self.delete_pod(self.get_pod_name('identity', self.config_attributes['managment_namespace']).rstrip())
            logging.info('Waiting for gateway service 30 seconds...')
            time.sleep(30)

            # Publish Live models and load cubes
            # ToDo leads to problem with identity replacement on multinode
            # api = self.api_authurization()
            # api.publish_and_load_cubes()

            return True
        except Exception as ex:
            return self.exception(ex)

    def mongodb_single_or_cluster_get_pods_name(self):
        try:
            cmd_single_or_cluster = '{0} -n {1} get nodes | grep -v NAME | wc -l'.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'])
            logging.info('mongodb_single_or_cluster_get_pods_name -> {0}'.format(cmd_single_or_cluster))
            single_or_cluster = int(subprocess.check_output(cmd_single_or_cluster, shell=True).decode('utf-8').split('\n')[0])
            mongodb_pod_name = None
            label = ''
            # Getting for L2021.9.0 and later
            try:
                label = 'app.kubernetes.io/name=mongodb'
                mongodb_container_name = 'mongodb'
                cmd_get_pod_name = '{0} -n {1} get pod -l {2}'.\
                                       format(self.kubectl_call, self.config_attributes['mongodb_namespace'], label)+ \
                                   ' -o jsonpath="{.items[0].metadata.name}"'
                logging.info('Get mongodb pod name -> {0}'.format(cmd_get_pod_name))
                mongodb_pod_name = subprocess.check_output(cmd_get_pod_name, shell=True).decode('utf-8')
                if single_or_cluster > 1:
                    members_command = "rs.status().members"
                    cmd_members = '{0} -n {1} -c {2} exec -i {3} -- mongo prismWebDB --eval \'{4}\''. \
                        format(self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name,
                               mongodb_pod_name, members_command)
                    logging.info('Get cluster mongodb members-> {0}'.format(cmd_members))
                    mongo_members = subprocess.check_output(cmd_members, shell=True).decode('utf-8')
                    logging.info(mongo_members)

                    eval_command = "rs.status().members.find(r=>r.state===1).name"
                    cmd_get_pod_name = '{0} -n {1} -c {2} exec -i {3} -- mongo prismWebDB --eval \'{4}\''. \
                        format(self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name,
                               mongodb_pod_name, eval_command)
                    logging.info('Get mongodb master pod name in cluster-> {0}'.format(cmd_get_pod_name))
                    mongodb_pod_name = \
                        subprocess.check_output(cmd_get_pod_name, shell=True).decode('utf-8').split('\n')[-2].split('.')[0]
                    logging.info('mongodb_pod_name -> {0} ; mongodb_container_name -> {1}'.
                                 format(mongodb_pod_name, mongodb_container_name))
                return mongodb_pod_name, mongodb_container_name
            except Exception as ex:
                logging.debug('Cannot get mongo pod name for label: {}. {}'.format(label, ex))

            if single_or_cluster == 1:
                # mongodb_pod_name = 'sisense-mongod-0'
                label = 'app=mongo'
                mongodb_container_name = 'mongod-container'
                cmd_get_pod_name = '{0} -n {1} get pod -l {2}'.\
                                       format(self.kubectl_call, self.config_attributes['mongodb_namespace'], label)+\
                                   ' -o jsonpath="{.items[0].metadata.name}"'
                logging.info('cmd_get_pod_name -> {0}'.format(cmd_get_pod_name))
                # Try for single node
                try:
                    mongodb_pod_name = subprocess.check_output(cmd_get_pod_name, shell=True).decode('utf-8')
                    logging.info('mongodb_pod_name -> {0} ; mongodb_container_name -> {1}'.format(mongodb_pod_name,
                                                                                                  mongodb_container_name))
                    return mongodb_pod_name, mongodb_container_name
                except Exception as exc:
                    logging.debug('Cannot get mongo pod name for label: {}. {}'.format(label, exc))

            # Cluster option get the primary pod name
            if not mongodb_pod_name:
                #label = 'app=mongodb-replicaset'
                mongodb_container_name = 'mongodb-replicaset'
                replica_pod_name = '{}-mongodb-replicaset-0'.format(self.config_attributes['mongodb_namespace'])
                eval_command = "rs.status().members.find(r=>r.state===1).name"
                cmd_get_pod_name = '{0} -n {1} -c {2} exec -i {3} -- mongo prismWebDB --eval \'{4}\''.\
                    format(self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name,
                           replica_pod_name, eval_command)
                logging.info('cmd_get_pod_name -> {0}'.format(cmd_get_pod_name))
                mongodb_pod_name = \
                subprocess.check_output(cmd_get_pod_name, shell=True).decode('utf-8').split('\n')[-2].split('.')[0]
                logging.info('mongodb_pod_name -> {0} ; mongodb_container_name -> {1}'.format(mongodb_pod_name,
                                                                                          mongodb_container_name))
                return mongodb_pod_name, mongodb_container_name
        except Exception as ex:
            return self.exception(ex)

    def mongodb_backup(self, pod_name, mongodb_container_name):
        try:
            logging.info('Running MongoDB dump')
            folder_name = 'mongodb_dump_original_' + str(time.strftime("%Y-%m-%d_%H_%M", time.gmtime()))
            cmd_mkdir = '{0} -n {1} -c {2} exec -i {3} -- mkdir tmp/{4}'.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'],
                       mongodb_container_name, pod_name, folder_name)
            self.continue_validation(self.execute_cmd(cmd_mkdir), 'cmd_mkdir', True)
            cmd_mongodump = '{0} -n {1} -c {2} exec -i {3} -- mongodump --out=tmp/{4}'.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'],
                       mongodb_container_name, pod_name, folder_name)
            self.continue_validation(self.execute_cmd(cmd_mongodump), 'cmd_mongodump', True)
            logging.info('Finished MongoDB dump')
            sub_backup_folder = os.path.join(self.config_attributes['working_dir'], folder_name)
            os.makedirs(sub_backup_folder)
            copy_cmd = '{0} cp -c {1} {2}/{3}:tmp/{4} {5}'.\
                format(self.kubectl_call, mongodb_container_name, self.config_attributes['mongodb_namespace'],
                       pod_name, folder_name, sub_backup_folder)
            self.continue_validation(self.execute_cmd(copy_cmd), 'copy_cmd', True)
            return True
        except Exception as ex:
            return self.exception(ex)

    def migration_job_update(self):
        logging.info("------------------------------------Migration job update------------------------------------")
        # Get migration job name
        migration_job_name = self.execute_cmd('{0} -n {1} get jobs -l app.kubernetes.io/name=migration -o name'.format(
            self.kubectl_call, self.config_attributes['mongodb_namespace']))
        if not migration_job_name:
            k8s_jobs = self.execute_cmd('{0} -n {1} get jobs'.format(self.kubectl_call, self.config_attributes['mongodb_namespace']),
                                        print_out=False)
            logging.warning('Migration job is not found. Jobs list: {}'.format(k8s_jobs))
            return
        else:
            migration_job_name = migration_job_name.split('/')[-1].rstrip()
            logging.info('Migration job name: {}'.format(migration_job_name))

        migration_job_backup_path = os.path.join(self.config_attributes['working_dir'], migration_job_name)

        try:
            logging.info('Backup migration job before re-running')
            self.execute_cmd('{0} -n {1} get job "{2}" -o json > {3}'.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'], migration_job_name, migration_job_backup_path))
            restart_migration_job_cmd = '{0} -n {1} get job "{2}" -o json |  ' \
                                        'jq \'del(.metadata.uid)\' | jq \'del(.spec.selector)\' | ' \
                                        'jq \'del(.spec.template.metadata.labels)\' | kubectl replace --force -f - '.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'], migration_job_name)
            self.continue_validation(self.execute_cmd(restart_migration_job_cmd), 'migration_job_update', True)
            return
        except Exception as ex:
            logging.error('Failed to replace migration job, restoring from backup.')
            create_backup_migration_job_cmd = '{0} -n {1} apply -f {2}'.format(self.kubectl_call, self.config_attributes['mongodb_namespace'], migration_job_backup_path)
            self.execute_cmd(create_backup_migration_job_cmd)
            return self.exception(ex)

    def mongodb_cleanup(self):
        logging.info("------------------------------------Mongodb cleanup------------------------------------")
        cleanup = MongoDB()
        self.continue_validation(cleanup.update_users_collection_with_duplicates(), 'update_users_collection_with_duplicates', True)
        self.continue_validation(cleanup.remove_elasticubes_that_are_not_live(), 'remove_elasticubes_that_are_not_live', True)

    def update_mongodb_new_identity(self, linux_new_identity):
        try:
            mongodb_pod_name, mongodb_container_name = self.mongodb_single_or_cluster_get_pods_name()
            logging.info('Update mongodb server identity')
            eval_command = 'db.servers.update({"address": "LocalHost"}, {"$set": {"identity":"'+linux_new_identity+'"}}, {multi:true})'
            cmd = '{0} -n {1} -c {2} exec -i {3} -- mongo prismWebDB --eval \'{4}\''.\
                format(self.kubectl_call, self.config_attributes['mongodb_namespace'], mongodb_container_name, mongodb_pod_name, eval_command)
            self.execute_cmd(cmd)
            return True
        except Exception as ex:
            return self.exception(ex)

    def delete_pod(self, pod_name):
        try:
            logging.info("Restarting pod -> {0}".format(pod_name))
            cmd_delete_pod = "{0} -n {1} delete pod {2} --wait".format(self.kubectl_call, self.config_attributes['managment_namespace'], pod_name)
            self.continue_validation(self.execute_cmd(cmd_delete_pod), 'cmd_delete_pod', False)
            logging.info('Sleeping for 15 seconds...')
            time.sleep(15)
            return True
        except Exception as ex:
            return self.exception(ex)

    def replace_identity(self):
        logging.info("------------------------------------Replace identity------------------------------------")
        api = self.api_authurization()
        linux_new_identity = api.get_new_linux_server_identity()
        self.continue_validation(linux_new_identity, 'Get new server identity', True)
        self.continue_validation(self.update_mongodb_new_identity(linux_new_identity), 'Update server identity', True)

        # Restart graphql
        self.delete_pod(self.get_pod_name('model-graphql', self.config_attributes['managment_namespace']).rstrip())
        logging.info("Finished replacing server identity")

    def all_steps(self, assets_path):
        self.mongodb_restore(assets_path)
        self.copy_plugins_to_destination(assets_path)
        self.copy_static_assets_to_destination(assets_path)
        self.replace_identity()
        self.mongodb_cleanup()


class MongoDB(K8S):
    ''' Class for connecting MongoDB and clean duplicates that prevent system to work'''

    def __init__(self):
        super(MongoDB, self).__init__()
        self.namespace = self.config_attributes['mongodb_namespace']
        self.pod_name, self.mongodb_container_name = self.mongodb_single_or_cluster_get_pods_name()

    def db_query(self, db_name, query):
        return "{0} -n {1} exec -i {2} -c {3} -- mongo {4} --quiet --eval '{5}'".\
            format(self.kubectl_call, self.namespace, self.pod_name, self.mongodb_container_name, db_name, query)

    def get_user_with_role_id(self, role):
        try:
            # Find role id
            role_id_query = 'db.roles.find({name: "'+role+'"}, {_id:1}).forEach(function(results){results._id=results._id.toString();printjson(results)})'
            role_record = self.execute_cmd(self.db_query('prismConfig', role_id_query))
            if not role_record:
                logging.error('Cannot find role "{}"'.format(role))
                return None, None, None, None
            else:
                role_object = json.loads(role_record)
                role_id = role_object['_id']

            # Find user
            user_id_query = 'printjson(db.users.find({roleId: '+role_id+', active: true}, {_id:0, email:1, hash:1}).toArray())'
            user_record = self.execute_cmd(self.db_query('prismWebDB', user_id_query), print_out=False)
            user_val = eval(user_record)
            if not user_val:
                logging.error('Cannot find users with role "{}"'.format(role))
                return None, None, None, None
            else:
                return user_val[0]['email'], user_val[0]['hash']['hash'], user_val[0]['hash']['salt'], user_val[0]['hash']['type']
        except Exception as ex:
            self.exception(ex)

    def change_user_password(self, user_email, hash_hash, hash_salt, hash_type):
        try:
            update_query = 'db.users.updateOne({email: "'+user_email+'"}, {"$set": {"hash.hash": "'+hash_hash+'", "hash.salt": "'+hash_salt+'", "hash.type": "'+hash_type+'"}})'
            update_res = self.execute_cmd(self.db_query('prismWebDB', update_query))
            update_val = json.loads(update_res)
            if update_val['matchedCount']:
                return True
            else:
                return False
        except Exception as ex:
            self.exception(ex)

    def update_users_collection_with_duplicates(self):
        try:
            users_query = 'db.users.find({}, {_id: 0, email: 1, groups: 1}).toArray()'
            users_resp = self.execute_cmd(self.db_query('prismWebDB', users_query), print_out=False)
            collection_users = json.loads(users_resp)

            for i in collection_users:
                if 'groups' in i:
                    if self.any_duplicates(i['groups']):
                        logging.info('Update duplicates within email: {}'.format(i['email']))
                        new_groups = self.remove_duplicates_from_list(i['groups'])
                        update_query = 'db.users.updateOne({email: "'+i['email']+'"}, {"$set": {"groups": '+str(new_groups).replace('\'', '\"')+'}})'
                        update_res = self.execute_cmd(self.db_query('prismWebDB', update_query))
                        try:
                            update_val = json.loads(update_res)
                            if update_val['matchedCount']:
                                logging.info('Groups duplicate for user {} were updated'.format(i['email']))
                            else:
                                logging.warning('Groups duplicate for user {} were not updated'.format(i['email']))
                        except Exception:
                            logging.warning('Groups duplicate for user {} were not updated'.format(i['email']))
                    else:
                        continue
            return True
        except Exception as ex:
            self.exception(ex)
        logging.info('Finished removing duplicates groups from users')

    @staticmethod
    def remove_duplicates_from_list(l):
        return list(set(l))

    @staticmethod
    def any_duplicates(l):
        seen = set()
        for x in l:
            if x in seen:
                return True
            seen.add(x)
        return False

    def remove_elasticubes_that_are_not_live(self):
        try:
            cubes_query = 'db.elasticubes.find({"type": {$ne: "live"}, "server": {$nin: ["LocalHost", "localhost"]}}, {_id: 0, type: 1, server: 1, title: 1}).toArray()'
            cubes_resp = self.execute_cmd(self.db_query('prismWebDB', cubes_query), print_out=False)
            logging.debug('Cubes will be deleted: {}'.format(cubes_resp))
            delete_query = 'db.elasticubes.deleteMany({"type": {$ne: "live"}, "server": {$nin: ["LocalHost", "localhost"]}})'
            self.execute_cmd(self.db_query('prismWebDB', delete_query), print_out=True)

            return True
        except Exception as ex:
            self.exception(ex)
        logging.info('Finished removing elaticubes')


def init_parser():
    """ Init end user arguments """
    # Read config.json
    config_json_location = os.path.join(os.getcwd(), "config.json")
    config_attributes = dict()
    with open(config_json_location) as config_file:
        config = json.load(config_file)

    config_attributes['linux_base_url'] = config["LINUX_TOKEN_LOGIN"]["LINUX_BASE_URL"]
    config_attributes['login_urlsuffix'] = config["LINUX_TOKEN_LOGIN"]["URL_SUFFIX"]
    config_attributes['working_dir'] = os.getcwd()
    config_attributes['input_files_path'] = config["LINUX_INPUT_FILES_PATH"]
    config_attributes['input_files_arc_name'] = config["LOCAL_DATASOURCES_ARC_NAME"] + ".zip"

    config_attributes['ecdata_storage_path'] = config["LINUX_IMPORT_ECDATA"]["STORAGE_PATH"]
    config_attributes['managment_namespace'] = config["LINUX_IMPORT_ECDATA"]["NAMESPACE"]

    config_attributes['plugins'] = '/usr/src/app/plugins-service/src'
    config_attributes['static_assets_paths'] = {
        'translations': '/opt/sisense/storage/translations',
        'branding': '/opt/sisense/storage/branding',
        'resources': '/opt/sisense/storage/resources',
        'templates': '/opt/sisense/storage/emails'
    }

    config_attributes['mongodb_namespace'] = config["LINUX_MONGODB"]["NAMESPACE"]

    # Parse CMD options
    parent_parser = argparse.ArgumentParser(description="The parent parser")
    parent_parser.add_argument('--extract', help='Extract sisense_win_assets.tar.gz', default='True', choices=['True', 'False'])
    parent_parser.add_argument('--api_username', help='Sisense API username', default=config["SISENSE_API"]["USERNAME"])
    parent_parser.add_argument('--api_password', help='Sisense API password', default=config["SISENSE_API"]["PASSWORD"])
    parent_parser.add_argument('--api_token', help='Sisense API access token (instead of username and password)', default=config["SISENSE_API"]["TOKEN"])

    parser = argparse.ArgumentParser(prog='win2linux_migration', description='Get function to execute')
    subparsers = parser.add_subparsers(help='sub-command help', dest='node')

    subparsers.add_parser('extract', help='Extract sisense_win_assets.tar.gz')
    subparsers.add_parser('copy_to_storage', parents=[parent_parser], add_help=False, help='Copy assets to '
                                                                                           'opt/sisense/storage/ and /usr/src/app/plugins-service location. '
                                                                                           'Apply branding. '
                                                                                           'Extract local datasources (xlsx, csv)')
    subparsers.add_parser('mongodb_restore', parents=[parent_parser], add_help=False, help='MongoDB restore')
    subparsers.add_parser('replace_identity', parents=[parent_parser], add_help=False,
                          help='Replace server identity inside MongoDB collection')
    subparsers.add_parser('mongodb_cleanup', parents=[parent_parser], add_help=False, help='MongoDB Cleanup')
    subparsers.add_parser('all_steps', parents=[parent_parser], add_help=False,
                          help='Activate all steps: extract -> extract_local_datasources -> mongodb restore  -> '
                               'copy_to_storage -> replace_identity -> mongodb_cleanup')
    args = parser.parse_args()
    args_dict = vars(args)
    args_dict.update(config_attributes)
    return args_dict


def main():

    params_dict.update(init_parser())
    utils = Utils()
    k8s = K8S()

    logging.info('Running "{}" mode'.format(params_dict['node']))

    try:
        if params_dict['node'] == 'extract':
            params_dict['extract'] = 'True'
            utils.general_extraction()

        elif params_dict['node'] == 'copy_to_storage':
            assets_path = utils.general_extraction()
            k8s.copy_ecdata_files_to_destenation(assets_path)
            k8s.copy_plugins_to_destination(assets_path)
            k8s.copy_static_assets_to_destination(assets_path)
            k8s.extract_local_datasources(os.path.join(assets_path, params_dict['input_files_arc_name']),
                                          params_dict['input_files_path'])

        elif params_dict['node'] == 'mongodb_restore':
            assets_path = utils.general_extraction()
            k8s.mongodb_restore(assets_path)
            k8s.replace_identity()
            k8s.mongodb_cleanup()

        elif params_dict['node'] == 'replace_identity':
            k8s.replace_identity()

        elif params_dict['node'] == 'mongodb_cleanup':
            k8s.mongodb_cleanup()

        elif params_dict['node'] == 'all_steps':
            assets_path = utils.general_extraction()
            k8s.extract_local_datasources(os.path.join(assets_path, params_dict['input_files_arc_name']),
                                          params_dict['input_files_path'])
            k8s.all_steps(assets_path)

    except Exception as ex:
        logging.error('Error {0} occurred when running <{1}>'.format(ex, params_dict['node']))
        raise


if __name__ == '__main__':
    START = time.time()
    main()
    logging.info('Linux side migration completed! script took: {0} seconds'.format(str(time.time() - START)))
